/***************************************************************************
                          Monster.cpp  -  description
                             -------------------
    begin                : Wed May 21 2003
    copyright            : (C) 2003 by Tyler Montbriand
    email                : tsm@accesscomm.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "Monster.h"
#include "Hero.h"
#include "Game.h"

/*! \brief Odds of enemy's new behavior, when done hiding             */
extern int    HideBehaveOdds[];
/*! \brief New behaviors to choose from when done hiding              */
extern int     HideBehaveRes[];
/*! \brief Odds of enemy's new behavior, when done showing            */
extern int    ShowBehaveOdds[];
/*! \brief New behaviors to choose from when done showing             */
extern int     ShowBehaveRes[];
/*! \brief Odds of enemy's new behaviors, when done taking a potshot  */
extern int   PshotBehaveOdds[];
/*! \brief New behaviors to choose from when done taking a potshot    */
extern int    PshotBehaveRes[];

/*! \brief Average firing rate of an enemy when taking potshots */
#define TICKS_PER_POTSHOT 550
/*! \brief Average firing rate of an enemy when pursing the player */
#define TICKS_PER_SHOT 750

/**
 *  The default constructor sets the target, i.e. player sprite
 *  and resets it to a default state.
 */
Monster::Monster(int idex, GridSprite *target):GridSprite(0,0,MONSTER_SPEED)
{
  hero=target;
  phase=0;
  deathhunt=0;
  index=idex;
  Reset();
}

/**
 *  Enemies die on all collisions, but the player only scores points
 *  when they shot the enemy.  So the GAME_EVENT_MONSTERDIED event
 *  is only produced when the player shot an enemy.
 */
void Monster::OnCollide(const Sprite *spr, ColType index)
{
  switch(index)
  {
  case GAME_LAYER_PFIRE:
    PushUserEvent(GAME_EVENT_MONSTERDIED);
    /* Falls through to Die(); */      
  case GAME_LAYER_HERO:
  case GAME_LAYER_ENEMY:
    Die();
    break;
    
  default:
    break;
  }
}

/**
 *  Seek toward the player's new position
 */
void Monster::SeekPlayer()
{
  Seek(hero->GridX(),hero->GridY(),2);  
}

/**
 *  This is the function that determines an enemy's behavior.
 *  An enemy performs a certain behavior until it's finished
 *  moving AND the timer runs out, after which SwitchOdds is called
 *  to choose a new behavior.
 */
void Monster::Think(Sint32 dtick)
{
  /* Timer needs to be incremented, not based on SDL_GetTicks() */
  timer.Advance(dtick);
  
  switch(mode)
  {
  case BHV_HIDE:  /* Keep hidden until the timer completes */
    if(timer.Left(delay)) Seek(HomeX(),HomeY(),1);  /* Move into box */
    else                  SwitchBehavior();     /* Decide on new behavior */
    break;

  case BHV_SHOW:  /* Keep in the open until the timer completes */
    /* Don't switch until we've stopped moving */
    if( ( gridx!=OutX() ) || ( gridy!=OutY() )|| Direction() )
    {
      Seek(OutX(),OutY(),1);
    }
    else if(!timer.Left(delay))
    {
      SwitchBehavior();
    }    
    break;

  case BHV_POTSHOT: /* Take potshots until the behavior completes */
    if(timer.Left(delay))
    {
      if((rand()%TICKS_PER_POTSHOT)<dtick)
      {
        PushUserEvent(GAME_EVENT_EFIRE,this,(void *)PshotDir());        
      }
    }
    else
    {      
      SwitchBehavior();   /* Decide on new behavior */
    }
    break;

  case BHV_TOGRID:  /* Move right onto the game grid */
    /* Don't finish until completed moving into grid */
    if( (gridx!=ToGrdX())||(gridy!=ToGrdY()) || Direction())
    {
      Seek(ToGrdX(),ToGrdY(),2);
    }
    else
    {
      SwitchBehavior();
    }
    break;

  case BHV_HUNT:  /* Hunt the player down and shoot at him */
    SeekPlayer();
    if((rand()%TICKS_PER_SHOT)<dtick)
    {
      PushUserEvent(GAME_EVENT_EFIRE,this,(void *)Direction());
    }
    break;

  default:
    fprintf(stderr,"Invalid behavior %d\n",mode);
    Reset();
    break;
  }
}

/**
 *  Called to choose a new behavior when the current cycle
 *  completes.
 */
void Monster::SwitchBehavior()
{
  int nbehv=0;

  /*  If we've been given a GAME_EVENT_DEATHHUNT event, hunt
      down the player under all circumstances                 */
  if(deathhunt)
  {
    DeathHunt();
    return;
  }
    
  switch(mode)
  {
  case BHV_HIDE:  /* Can either stay hiding, or show itself */
    nbehv=SwitchOdds(HideBehaveOdds,HideBehaveRes);
    DoBehavior((Behavior_e)nbehv);
    break;

  case BHV_SHOW:  /* Can either hide, take potshot, or hunt */
    nbehv=SwitchOdds(ShowBehaveOdds,ShowBehaveRes);
    DoBehavior((Behavior_e)nbehv);
    break;

  case BHV_POTSHOT: /* Can either take a potshot, hunt, or hide */
    nbehv=SwitchOdds(PshotBehaveOdds,PshotBehaveRes);
    DoBehavior((Behavior_e)nbehv);  
    break;

  case BHV_TOGRID:  /*  Hunt down player */
    Hunt();
    break;
    
  default:
    break;
  }
}

/**
 *  Put an explosion where the enemy was, put it back in it's
 *  starting position, and advance one phase.
 *
 *  How the phases work is, horizontal rows of frames are
 *  displayed sequentially...  a 4*4 sprite image would have
 *  4 phases, phase 0 using frames 0 thru 3, phase 1 using
 *  4 through 7, etc.  Once we reach phase 4, we're out of frames,
 *  and we tell the sprite to die.
 */
void Monster::Die()
{
  PushUserEvent(GAME_EVENT_EXPLOSION,(void *)Xfxp(),(void *)Yfxp());
    
  Reset();

  /* If we've run out of horizontal rows of frames, we're done. */
  if(phase>=(frames.y-1))
  {
    PushUserEvent(GAME_EVENT_MONSTERGONE);  /* Notify game of empty slot */
    Expire(); /* Remove from game */
  }
  else
  { 
    phase++;      /* Lets go next phase */
    CalcFrame();  /* Calculate new frame position */
  }
}

/**
 *  This is enemy sprite's handler for SDL_USEREVENTs.
 */
void Monster::Event(int code, void *data1, void *data2)
{
  switch(code)
  {
  case GAME_EVENT_DEATHHUNT:  /* Causes all enemies to hunt down the player */
    deathhunt=1;
    break;
    
  case GAME_EVENT_PLAYERDIED: /* Causes all enemies to return to start pos  */
    Reset();
    break;

  case GAME_EVENT_NUKE: /*  Kills all enemies that aren't hiding */
    if(mode!=BHV_HIDE)  Die();
    break;

  default:
    break;
  }
}

/**
 *  Instead of proceeding from first frame to last frame
 *  in linear order, it repeats individual rows of frames.
 *  'frame' is the current column, and 'phase' is the row.
 */
int Monster::GetFrame()
{
  int f=(FROM_SFXP(frame)%frames.x)+(phase*frames.x);
  return(f);
}

/**
 *  When in deathhunt mode, this function handles state changes
 *  with enemies going from hiding to hunting as directly as
 *  possible, instead of behavior being chosen randomly.
 */
void Monster::DeathHunt()
{
  switch(mode)
  {
  case BHV_HIDE:    /* When done hiding, go direct to showing         */
    Show();
    break;

  case BHV_SHOW:    /* When done showing or potshotting, go to grid */
  case BHV_POTSHOT:
    /*  Note - this could be simplified since BHV_SHOW, BHV_POTSHOT,
        and BHV_TOGRID all trigger the same function call, Hunt().
        But Hunt() triggers the BHV_TOGRID state when called in
        BHV_SHOW or BHV_POTSHOT states, and the BHV_HUNT state when
        called from the BHV_TOGRID state, so the two cases have been
        left for clarity.                                           */
    Hunt();
    break;

  case BHV_TOGRID:  /* When done moving to grid, go direct to hunting */
    Hunt();
    break;

  case BHV_HUNT:
    fprintf(stderr,"Warning - switching to BHV_HUNT from BHV_HUNT, why?\n");
    break;
  }
}

/**
 *  Sets behavior to hiding, resets the timer, and generates a random delay
 */
void Monster::Hide()
{
  mode=BHV_HIDE;
  delay=RandRange(DELAY_HIDE_MIN,DELAY_HIDE_MAX);
  timer.Reset();
}

/**
 *  Sets behavior to showing, resets the timer, and generates a random delay
 */
void Monster::Show()
{
  mode=BHV_SHOW;
  delay=RandRange(DELAY_SHOW_MIN,DELAY_SHOW_MAX);
  timer.Reset();
}

/**
 *  Sets behavior to potshot, resets the timer, and generates a random delay
 */
void Monster::Potshot()
{
  mode=BHV_POTSHOT;
  delay=RandRange(DELAY_PSHOT_MIN,DELAY_PSHOT_MAX);
  timer.Reset();
}

/**
 *  Sets behavior to moving to grid.  This phase
 *  doesn't use delay or timer.
 */
void Monster::MoveToGrid()
{
  mode=BHV_TOGRID;
}

/**
 *  Sets behavior to either moving to grid or hunting.
 *  These phases don't use delay or timer.
 */
void Monster::Hunt()
{
  if(mode!=BHV_TOGRID)
  {
    mode=BHV_TOGRID;
  }
  else
  {
    mode=BHV_HUNT;
  }
}

/**
 *  Switches to the specified state.
 */
void Monster::DoBehavior(Behavior_e beh)
{
  switch(beh)
  {
  case BHV_HIDE:
    Hide();
    break;

  case BHV_SHOW:
    Show();
    break;

  case BHV_POTSHOT:
    Potshot();
    break;

  case BHV_TOGRID:
    MoveToGrid();
    break;

  case BHV_HUNT:
    Hunt();
    break;

  default:
    fprintf(stderr,"Monster::DoBehavior(%d), unknown behavior\n",beh);
    break;
  }
}


/**
 *  Seeks toward the specified x/y grid coordinates.
 *
 *  So the enemies don't ALL run into each other
 *  when seeking towards the player, they take
 *  2 different routes:  Even-numbered enemies
 *  seek in the X direction first, while
 *  odd numbered ones seek in the Y direction first.
 */
void Monster::Seek(int gx, int gy, int step)
{
  int dir=Direction();

  if(!dir)
  {
    if(index&1) /* Approach from top, or side? */
    {
      if(gridx>gx)      gridx-=step;
      else if(gridx<gx) gridx+=step;
      else if(gridy>gy) gridy-=step;
      else if(gridy<gy) gridy+=step;
    }
    else
    {
      if(gridy>gy)      gridy-=step;
      else if(gridy<gy) gridy+=step;
      else if(gridx>gx) gridx-=step;
      else if(gridx<gx) gridx+=step;
    }
  }
  else
  {
    if(index&1) /* Approach from top, or side? */
    {
      if(dir&DIR_LEFT)        if(gridx<gx)  gridx+=step;
      else if(dir&DIR_RIGHT)  if(gridx>gx)  gridx-=step;
      else if(dir&DIR_DOWN)   if(gridy>gy)  gridy-=step;
      else if(dir&DIR_UP)     if(gridy<gy)  gridy+=step;
    }
    else
    {
      if(dir&DIR_DOWN)        if(gridy>gy)  gridy-=step;
      else if(dir&DIR_UP)     if(gridy<gy)  gridy+=step;
      else if(dir&DIR_LEFT)   if(gridx<gx)  gridx+=step;
      else if(dir&DIR_RIGHT)  if(gridx>gx)  gridx-=step;
    }
  }
}

/**
 *  Overloaded update function that calls the AI
 *  routines every game tick
 */
void Monster::Update(Sint32 dtick)
{
  Think(dtick);
  GridSprite::Update(dtick);
}

void Monster::Reset()
{
  Hide();
  SnapTo(HomeX(),HomeY());
}

/**
 *  Calculates the hiding, showing, move-to-grid coordinates, and
 *  firing directions for all enemies.  Works for arbitrary grid sizes
 *  with the condition that GCOUNT_X and GCOUNT_Y must be odd numbers.
 *
 *  How it works is this:  It starts in the upper left hand corner,
 *  moving right, placing sprite positions in a line.  When it hits
 *  the right side it changes direction, moving downwards.  On hitting
 *  the bottom, it starts moving left, and when it hits the left
 *  side it goes upwards, completing the circle.  If it
 *  hits the top again(it shouldn't unless MAX_MONSTERS is wrong somehow),
 *  it breaks the loop.
 */
void calc_monster_pos()
{
  /* Sprite positions */
  int gx=1,gy=0;
  /* Direction to increment position is */
  int dir_line=DIR_RIGHT;
  /* Loop counter */
  int n=0;


  for(n=0; n<MAX_MONSTERS; n++)
  {
    /* Set initial position */
    monster_in_x[n]=gx;
    monster_in_y[n]=gy;
    /* Set outside position equal to initial position, for now */
    monster_out_x[n]=gx;
    monster_out_y[n]=gy;
    /*  Increment the outside position one grid
        element away in the direction we're going */
    Offset(dir_line,1,monster_out_x[n],monster_out_y[n]);
    /*  Set the grid position equal to the outside position, for now */
    monster_grd_x[n]=monster_out_x[n];
    monster_grd_y[n]=monster_out_y[n];
    /*  Increment the grid position 2 grid units and 90 degrees
        clockwise to the direction we're incrementing in, onto
        the grid, like so:   ^
                             |-->
     */
    Offset(RotateDir(dir_line),2,monster_grd_x[n],monster_grd_y[n]);
    /*  Set the firing direction 90 degrees clockwise
        to the direction we're running in             */
    monster_pshot_dir[n]=RotateDir(dir_line);


    if(dir_line&DIR_RIGHT) /*If we're moving right */
    {
      if((gx+2)==(GCOUNT_X-2))
      {
        /* If we've reached the end, start travelling down */
        gx+=3;
        gy+=1;
        dir_line=RotateDir(dir_line);
      }
      else
      {
        /* If not, just plain move right */
        gx+=2;
      }
    }
    else if(dir_line&DIR_DOWN)  /* If we're moving downwards */
    {
      if((gy+2)==(GCOUNT_Y-2))
      {
        /* If we've reached the end, start moving left */
        gx-=1;
        gy+=3;
        dir_line=RotateDir(dir_line);
      }
      else
      {
        /* If not, just plain move down */
        gy+=2;
      }
    }
    else if(dir_line&DIR_LEFT) /* If we're travelling left */
    {
      if((gx-2)==1)
      { /* If we've reached the end, start moving upwards */
        gx-=3;
        gy-=1;
        dir_line=RotateDir(dir_line);
      }
      else
      {
        /*Otherwise, just plain move left */
        gx-=2;
      }
    }
    else if(dir_line&DIR_UP)  /* If we're travelling upwards */
    {
      if((gy-2)==1)
      {
        /* If we've reached the end, we've
           gone full circle.  Break the loop.   */
        break;
      }
      else
      {
        /* Otherwise, continue moving up. */
        gy-=2;
      }
    }
  }
}



int PshotBehaveOdds[]={          2,       1,       2,-1};
int  PshotBehaveRes[]={BHV_POTSHOT,BHV_HUNT,BHV_HIDE};

int HideBehaveOdds[]={       1,       1, -1 };
int  HideBehaveRes[]={BHV_HIDE,BHV_SHOW     };
int ShowBehaveOdds[]={       2,          4,       1, -1 };
int  ShowBehaveRes[]={BHV_HIDE,BHV_POTSHOT,BHV_HUNT     };

/**
 *  Since every enemy has the same player as a target,
 *  they share the same static hero variable.
 */
GridSprite *Monster::hero=NULL;
int monster_pshot_dir[MAX_MONSTERS];
                          
int monster_grd_x[MAX_MONSTERS];
int monster_grd_y[MAX_MONSTERS];
int monster_out_x[MAX_MONSTERS];
int monster_out_y[MAX_MONSTERS];
int monster_in_x[MAX_MONSTERS];
int monster_in_y[MAX_MONSTERS];
