/***************************************************************************
                          Music.cpp  -  description
                             -------------------
    begin                : Sat May 10 2003
    copyright            : (C) 2003 by Tyler Montbriand
    email                : tsm@accesscomm.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include "Music.h"

/**
 *  Status variable, 0 when uninitialized, nonzero when initialized.
 *  It works recursively;  if it's initialized more than once,
 *  it gets incremented every time, and thus stays initialized
 *  until it's been de-initialized the same number of times.
 */
int        Music::music_init   = 0;
/**
 *  Contains the data representing the currently playing
 *  song, or NULL if none.
 */
Mix_Music *Music::music_data   = NULL;
/**
 *  This mutex is locked when music operations
 *  are performed, preventing multithreaded programs
 *  from attempting to access the music object
 *  simultaneously.
 */
SDL_mutex *Music::lock         = NULL;
/**
 *  The single, static music object that is used by everyone.
 */
Music      Music::music_object = Music();

/**
 *  Attempts to load, then begin playing the specified
 *  music file.  Returns nonzero on success, zero on failure.
 */
int Music::Play(const char *fname, int loops, int fadein, int pos)
{
  /* Can't play music if we're not initialized. */
  if(!music_init) return(0);
    
  Lock(); /*Lock the mutex to prevent simultaneous access */
  
  if(!Load(fname))  /* Load song  */
  {
    /* If couldn't load, return faliure */
    Unlock();
    return(0);
  }
  /* Attempt to play the newly-loaded song  */
  else if(Mix_FadeInMusicPos(music_data,loops,fadein,pos)<0)
  {
    /*If couldn't play, unload and return failure */
    Free();
    Unlock();
    return(0);
  }
  else
  {
    /* Return success */
    Unlock();
    return(1);
  }
}

/**
 *  Attempts to halt and unload the currently playing song.
 */
int Music::Stop()
{
  if(!music_init) return(0);
    
  Lock(); /* Lock mutex to prevent simultaneous access  */
  if(Playing()) /* If music is playing  */
  { 
    Mix_HaltMusic();  /*  Stop music        */
    Free();           /*  Free loaded song  */
    Unlock();
    return(1);        /*  Return success    */
  }
  else
  {
    Unlock();
    return(0);        /*  Return failure    */
  }
}

int Music::Playing()
{
  /* Obviously nothing's playing if we're not initialized yet */
  if(!music_init) return(0);
          
  Lock(); /* Lock mutex to prevent simultaneous access  */
  int retval=Mix_PlayingMusic();
  Unlock();
  
  return(retval);
}

int Music::Loaded()
{  
  if(!music_init) return(0);  /* Nothing can be loaded
                                  if we're not initialized  */
  else            return(music_data!=NULL);
}

/**
 *  Also halts any currently playing music.
 */
int Music::Load(const char *fname)
{
  if(!music_init) return(0);

  if(Loaded()) Free();  /* Stop any currently loaded/playing music  */

  music_data=Mix_LoadMUS(fname);  /* Attempt to load new music  */
  if(!Loaded()) /* If we couldn't load it, return failure */
  {
    return(0);
  }
  else  /* Return success */
  {
    return(1);
  }
}


int Music::Free()
{
  /* There's nothing to be freed if we're not initialized */
  if(!music_init) return(0);  
    
  if(Loaded())  /* if there is music loaded/playing */
  {
    Mix_HaltMusic();            /* Stop playing             */
    Mix_FreeMusic(music_data);  /* Free current music       */
    music_data=NULL;            /* Mark music as not loaded */
    return(1);                  /* Return success           */
  }
  else
  {
    return(0);                  /* Return failure           */
  }
}

/**
 *  It works recursively, like GetObject().  It doesn't actually
 *  deinitialize the music object until ~Music()'s been called
 *  the same number of times as GetObject()
 */
Music::~Music()
{
  /* What happens depends on number of instances being used.
     If we're getting rid of the last instance, deinitialize
     everything.  Otherwise leave it alone.                  */
  if(music_init>1)
  {
    music_init--;
  }
  else if(music_init==1)
  {
    if(Loaded()) Stop();
    SDL_DestroyMutex(lock);
    lock=NULL;
    music_init=0;
  }
  else if(music_init<0)
  {
    fprintf(stderr,"Music::~Music() Destructor called too many times!\n");
    fprintf(stderr,"Music::music_init == %d\n",music_init);
  }
}


Music::Music() { }
